#

__author__  = 'Scott Little'
__email__   = 'slittle@slittle.com'
__program__ = 'NodelistWidgets.py'
__version__ = '20120715'


class NodelistPubDates:
	from datetime import date, timedelta

	def __init__(self, year=None, pubday=4, fdate=None, ldate=None):
		self._now = self.date(2000, 1, 1).today()
		self._day = self.timedelta(days=1)

		if year:
			fdate = self.date(year, 1, 1)
			ldate = self.date(year+1, 1, 1) - self._day

		if not fdate:
			fdate = self.date(self._now.year, 1, 1)

		if not ldate:
			ldate = self.date(self._now.year+1, 1, 1) - self._day

		assert(isinstance(fdate, self.date))
		assert(isinstance(ldate, self.date))

		while fdate.weekday() != pubday:
			fdate += self._day

		while ldate.weekday() != pubday:
			ldate -= self._day

		self.fdate  = fdate
		self.ldate  = ldate
		self.pubday = pubday

	def __iter__(self):
		work = self.fdate

		while work <= self.ldate:
			tt = work.timetuple()
			yield (tt[7], work)
			work += self._day * 7

	def current(self):
		work = self._now

		while work.weekday() != self.pubday:
			work -= self._day

		return (work.timetuple()[7], work)

	def next(self):
		work = self._now + self._day

		while work.weekday() != self.pubday:
			work += self._day

		return (work.timetuple()[7], work)


def parse_header(buff):
	# ";A Zone Nodelist for Friday, January 29, 2010 -- Day number 029 : 48353"

	from datetime import date

	Months = dict(zip(('january', 'february', 'march', 'april', 'may', 'june',
	          'july', 'august', 'september', 'october', 'november', 'december'), range(1, 13)))
	Days = dict(zip(('monday', 'tuesday', 'wednesday', 'thursday', 'friday', 'saturday', 'sunday'), range(0, 7)))

	if '\r\n' in buff:
		buff = buff[:buff.find('\r\n')]

	crap, network, crap, crap, wday, month, day, year, crap, crap, crap, julian, crap, crc = buff.strip().split()
	wday = Days[wday.lower().strip(',')]
	month = Months[month.lower()]
	day = int(day.strip(','))
	year = int(year)
	julian = int(julian)
	crc = int(crc)
	return {'network': network, 'wday': wday, 'year': year, 'month': month, 'day': day, 'julian': julian, 'crc': crc, 'date': date(year, month, day)}


def regen_header(buff):
	# recalculate CRC on existing nodelist header
	header, body = buff.split('\r\n', 1)
	return '%s %s\r\n%s' % (header[:header.rfind(':')+1], calc_crc(buff), body)


class CRC_CCITT:
	def __init__(self):
		self._init_table()

	def _init_table(self):
		self.tab=256*[[]]
		for i in xrange(256):
			crc=0
			c = i << 8
			for j in xrange(8):
				assert(j == j) # pychecker bait
				if (crc ^ c) & 0x8000:
					crc = ( crc << 1) ^ 0x1021
				else:
					crc = crc << 1
				c = c << 1
				crc = crc & 0xffff
			self.tab[i]=crc

	def calc(self, buff):
		crcval = 0
		for x in buff:
			crcval = self.tab[((crcval >> 8) & 0xFF) ^ ord(x)] ^ (crcval << 8) & 0xFFFF
		return crcval


def calc_crc(buff):
	# expects the full nodelist, including header
	if buff.find('\x1A') > -1:
		return CRC_CCITT().calc(buff[buff.find('\r\n')+2:buff.find('\x1A')])
	else:
		return CRC_CCITT().calc(buff[buff.find('\r\n')+2:])


def test():
	import sys
	stdout = sys.stdout
	#stderr = sys.stderr

	p = NodelistPubDates(year=2000)

	stdout.write('Pubdates between %s and %s:\n' % (p.fdate, p.ldate))
	for q in p:
		stdout.write('  %s  %03s\n' % q[::-1])

	stdout.write('\n')

	stdout.write('Current edition:\n  %s  %03s\n\n' % p.current()[::-1])
	stdout.write('Next edition:\n  %s  %03s\n' % p.next()[::-1])

	stdout.write('\nHeader fields:\n')
	stdout.write('  %s\n' % parse_header(';A Zone Nodelist for Friday, January 29, 2010 -- Day number 029 : 48353'))


def main():
	import sys, optparse

	parser = optparse.OptionParser()
	parser.add_option('-t', '--test', action='store_true', dest='test', default=False, help='Run test calculations')
	parser.add_option('-d', '--date', dest='date', metavar='FILE', default=None, help='Show pubdate of specified nodelist')
	parser.add_option('-u', '--update', dest='update', metavar='FILE', default=None, help='Regenerate the CRC for the specified nodelist')
	opts, args = parser.parse_args()

	if opts.test:
		test()
		sys.exit(0)

	else:
		if opts.date:
			try:
				header = parse_header(file(opts.date, 'rb').read(4096))
			except:
				sys.exit(1)

			print('ISO: %s' % header['date'])
			print('Julian: %03i' % header['julian'])

		if opts.update:
			import os, shutil

			tmpname = opts.update + '.$tmp$'

			if os.path.exists(tmpname):
				print('ERROR: temporary file exists: %s' % tmpname)
				sys.exit(1)

			try:
				buff = file(opts.update, 'rb').read()
			except:
				print('FATAL: unable to read file: %s' % opts.update)
				sys.exit(1)

			buff = regen_header(buff)

			try:
				shutil.copy2(opts.update, tmpname)
				file(opts.update, 'wb').write(buff)
				os.remove(tmpname)
			except:
				print('FATAL: error writing updated file: %s' % opts.update)
				sys.exit(1)


if __name__ == '__main__':
	print('%s v%s by %s\n' % (__program__, __version__, __author__))
	main()
