#!/usr/bin/python


#	PktWidgets.py
#	Copyright (C) 2004  Scott Little, slittle@slittle.com


__author__	= "Scott Little"
__version__	= "20041223"
__program__	= "PktWidgets.py"



import time, struct

class Type2Plus:

	"Python container for Type2+ PKTs"


	class Error(StandardError):
		"""
		Catchall exception class.
		"""


	def __init__(self, PKTbuff=None, DefaultZone=1):

		self.DefaultZone = DefaultZone
		self.Header      = self.UnpackHeader(PKTbuff)
		self.Messages    = self.UnpackMessages(PKTbuff)



	def __str__(self):

		return self.PackHeader(self.Header) + self.PackMessages(self.Messages)



	def BlankHeader(self):
		"""
		Return an empty PKT header dictionary.
		"""

		now = time.localtime()

		return {
			'origzone2': self.DefaultZone,
			'destzone2': self.DefaultZone,
			'origzone' : self.DefaultZone,
			'destzone' : self.DefaultZone,
			'orignet'  : 0, 'orignode' : 0,	'origpoint': 0,
			'destnet'  : 0,	'destnode' : 0,	'destpoint': 0,
			'year'     : now[0], 'month' : now[1], 'day'   : now[2],
			'hour'     : now[3], 'minute': now[4], 'second': now[5],
			'baud'     : 0,
			'pkttype'  : 2,
			'prodcode1': 255, 'prodcode2': 0,
			'prodrev1' : 0,	'prodrev2' : 0,
			'password' : '',
			'auxnet'   : 0,
			'cw'       : 1,	'cw2'      : 1,
			'filler'   : '' }


	def BlankMessage(self):
		"""
		Return an empty PKT message dictionary.
		"""

		return {
			'type'    : 2,
			'orignode': 0,
			'destnode': 0,
			'orignet' : 0,
			'destnet' : 0,
			'attrib'  : 0,
			'cost'    : 0,
			'time'    : time.strftime('%d %b %y  %H:%M:%S', time.localtime()),
			'from'    : '',
			'to'      : '',
			'subject' : '',
			'body'    : ''  }


	def UnpackHeader(self, PKTbuff=None):

		Header = self.BlankHeader()

		if PKTbuff == None:
			return Header

		if len(PKTbuff) < 58:
			raise Type2Plus.Error("PKT format is not recognised")

		Header['orignode'],  \
		Header['destnode'],  \
		Header['year'],      \
		Header['month'],     \
		Header['day'],       \
		Header['hour'],      \
		Header['minute'],    \
		Header['second'],    \
		Header['baud'],      \
		Header['pkttype'],   \
		Header['orignet'],   \
		Header['destnet'],   \
		Header['prodcode1'], \
		Header['prodrev1'],  \
		Header['password'],  \
		Header['origzone'],  \
		Header['destzone'],  \
		Header['auxnet'],    \
		Header['cw2'],       \
		Header['prodcode2'], \
		Header['prodrev2'],  \
		Header['cw'],        \
		Header['origzone2'], \
		Header['destzone2'], \
		Header['origpoint'], \
		Header['destpoint'], \
		Header['filler'] =   \
			struct.unpack('<12H2B8s3H', PKTbuff[:40]) +  struct.unpack('>H', PKTbuff[40:42]) + struct.unpack('<2B5H4s', PKTbuff[42:58])

		# Verify Type-2+ packet
		if Header['pkttype'] != 2:
			raise Type2Plus.Error("PKT format not supported: %s" % Header['pkttype'])
		if Header['cw'] != Header['cw2']:
			raise Type2Plus.Error("PKT format not supported: CompatibilityWord does not match validation copy")

		return Header



	def UnpackMessages(self, PKTbuff=None):

		if PKTbuff == None:
			return []

		PKTbuff = PKTbuff[58:]
		messages = []

		while len(PKTbuff) > 38:
			message = self.BlankMessage()

			message['type'],     \
			message['orignode'], \
			message['destnode'], \
			message['orignet'],  \
			message['destnet'],  \
			message['attrib'],   \
			message['cost'],     \
			message['time'] =    \
				struct.unpack('<7H20s', PKTbuff[0:34])

			message['to'],     \
			message['from'],       \
			message['subject'],  \
			message['body'],               \
			PKTbuff =                     \
				PKTbuff[34:].split('\x00', 4)

			messages.append(message)

		return messages



	def PackHeader(self, header):

		return struct.pack('<12H2B8s3H',
			header['orignode'],
			header['destnode'],
			header['year'],
			header['month'],
			header['day'],
			header['hour'],
			header['minute'],
			header['second'],
			header['baud'],
			header['pkttype'],
			header['orignet'],
			header['destnet'],
			header['prodcode1'],
			header['prodrev1'],
			header['password'],
			header['origzone'],
			header['destzone'],
			header['auxnet']) + struct.pack('>H',
			header['cw2']) + struct.pack('<2B5H4s',
			header['prodcode2'],
			header['prodrev2'],
			header['cw'],
			header['origzone2'],
			header['destzone2'],
			header['origpoint'],
			header['destpoint'],
			header['filler'])



	def PackMessages(self, messages):

		buff = ''

		for message in messages:
			buff += struct.pack('<7h20s',
				message['type'],
				message['orignode'],
				message['destnode'],
				message['orignet'],
				message['destnet'],
				message['attrib'],
				message['cost'],
				message['time'])

			buff += '%(to)s\x00%(from)s\x00%(subject)s\x00%(body)s\x00' % message

		buff += chr(0)*2

		return buff



	def PrettyPrint(self, show_header=True, show_messages=True):

		"""
		Display packet contents formatted nicely.
		"""


		if show_header:
			print "="*75
			print "Type: %(pkttype)i" % self.Header
			print "Date: %s-%s-%s %s:%s:%s" % (self.Header['year'], str(self.Header['month']).zfill(2), str(self.Header['day']).zfill(2), str(self.Header['hour']).zfill(2), str(self.Header['minute']).zfill(2), str(self.Header['second']).zfill(2), )
			print "Prod: %s%s" % (hex(self.Header['prodcode2'])[2:].upper(), hex(self.Header['prodcode1'])[2:].upper())
			print "Ver:  %(prodrev1)i.%(prodrev2)i" % self.Header
			print "Orig: %(origzone)i:%(orignet)i/%(orignode)i.%(origpoint)i" % self.Header
			print "Dest: %(destzone)i:%(destnet)i/%(destnode)i.%(destpoint)i" % self.Header
			print "Pass: %(password)s" % self.Header
			print "="*75


		if show_header and show_messages:
			print

		if show_messages:
			count = 0

			for message in self.Messages:
				count += 1

				title = "[ %i of %i ]" % (count, len(self.Messages))

				while len(title) < 75:
					if (len(title) % 2) == 0:
						title += '='
					else:
						title = '=' + title

				print title
				#print "Msg:    %i of %i" % (count, len(self.Messages))
				print "From:   %(from)s, %(orignode)i/%(orignet)i" % message
				print "To:     %(to)s, %(destnode)i/%(destnet)i" % message
				print "Date:   %(time)s" % message
				print "Attrib:",

				if (message['attrib'] & (1 * pow(2, 0))) != 0:
					print "Pvt",
				if (message['attrib'] & (1 * pow(2, 1))) != 0:
					print "Crash",
				if (message['attrib'] & (1 * pow(2, 2))) != 0:
					print "Recd",
				if (message['attrib'] & (1 * pow(2, 3))) != 0:
					print "Sent",
				if (message['attrib'] & (1 * pow(2, 4))) != 0:
					print "FAttach",
				if (message['attrib'] & (1 * pow(2, 5))) != 0:
					print "InTransit",
				if (message['attrib'] & (1 * pow(2, 6))) != 0:
					print "Orphan",
				if (message['attrib'] & (1 * pow(2, 7))) != 0:
					print "Kill/Sent",
				if (message['attrib'] & (1 * pow(2, 8))) != 0:
					print "Local",
				if (message['attrib'] & (1 * pow(2, 9))) != 0:
					print "Hold",
				if (message['attrib'] & (1 * pow(2, 10))) != 0:
					print "<10>",
				if (message['attrib'] & (1 * pow(2, 11))) != 0:
					print "Freq",
				if (message['attrib'] & (1 * pow(2, 12))) != 0:
					print "RRR",
				if (message['attrib'] & (1 * pow(2, 13))) != 0:
					print "Rcpt",
				if (message['attrib'] & (1 * pow(2, 14))) != 0:
					print "AuditReq",
				if (message['attrib'] & (1 * pow(2, 15))) != 0:
					print "FUpd",
				print

				print "-"*75
				for line in message['body'].split('\r'):
					print line
				print "="*75
				print

				
if __name__ == '__main__':
	print "%s v%s by %s" % (__program__, __version__, __author__)

	#import psyco
	#psyco.full()

	import time
	StartTime = time.time()


	#pkt = Type2Plus()
	#file('test.out', 'wb').write(pkt.PackHeader(pkt.Header) + pkt.PackMessages(pkt.Messages))


	buff = file('test.pkt', 'rb').read()
	#buff = file('test2.pkt', 'rb').read()
	pkt = Type2Plus(buff)

	pkt.PrettyPrint()


	buff2 = pkt.PackHeader(pkt.Header) + pkt.PackMessages(pkt.Messages)
	
	if buff[0:60] != buff2[0:60]:
		print "meh"
	
	file('test.out', 'wb').write(buff2)


	EndTime = time.time()
	print "Runtime:", EndTime - StartTime
