#!/usr/bin/env python


#	Check segment for common problems:
#		* Validate header CRC (mandatory)
#		* Determine segment age (optional)
#		* Confirm presence/absence of nodes/masks (mandatory)
#		* XXXTODO: Require minimum/maximum number of nodes (optional)
#		* XXXTODO: message templates
#		* XXXTODO: semaphore path and/or [segment].fail (and [segment].warn?)
#		* Save successful segment (uncompressed) to specified path (optional)
#		* Exit with non-zero errorlevel on any problems
#		* Create semaphore on any problems (because errorlevel doesn't work through a pipe)
#		* Report results via PKT (optional)

#	To be called from catch_z3list, catch_zsegs, watch_regions and updatenl (before processing segments).


__author__  = 'Scott Little'
__email__   = 'slittle@slittle.com'
__program__ = 'check_segment.py'
__version__ = '20130924'


import sys, os, time, optparse, fnmatch, zipfile, uuid
import FTNAddress2, NodelistWidgets, PktWidgets


class NoClass:
	pass


## XXXFIXME: import read_nodelist from whatever module I had it in before
def read_nodelist(buff, addr):
	SLFTYPE_DOMAIN  = ('domain', )
	SLFTYPE_ZONE    = ('zone', )
	SLFTYPE_REGION  = ('region', )  # 'zone'
	SLFTYPE_NET     = ('host', )  # 'zone', 'region'
	SLFTYPE_HUB     = ('hub', )  # node
	SLFTYPE_NODE    = ('', 'node', 'hold', 'down', 'pvt')
	SLFTYPE_POINT   = ('point', )
	SLFTYPE_BOSS    = ('boss', 'bossnode')
	SLFTYPE_IGNORE  = ('down', 'kenl', 'error')

	for line, text in enumerate(buff):  # buff must be an iterable of strings with no line endings
		if (not text) or (text[0] == ';'):
			continue
		elif text[0] == '\x1a':
			break

		# split off the main (non-flag) fields
		bits = text.split(',', 7)
		if len(bits) < 2:
			raise UserWarning('FATAL: Invalid entry on line #%s; expecting at least two fields.' % line)


		node = NoClass()
		node.line = line
		node.text = text
		node.addr = addr  # XXXWARNING: for speed this structure is yield'ed directly, not copied.  ie. don't change it.
		node.type = bits[0].lower()

		if not node.type in SLFTYPE_DOMAIN + SLFTYPE_ZONE + SLFTYPE_REGION + SLFTYPE_NET + SLFTYPE_HUB + SLFTYPE_NODE + SLFTYPE_POINT + SLFTYPE_BOSS + SLFTYPE_IGNORE:
			raise UserWarning('FATAL: Invalid entry on line #%s; unrecognised node type.' % line)

		if node.type in SLFTYPE_IGNORE:
			continue

		if node.type in SLFTYPE_BOSS:
			if not addr.parse_5d(bits[1]):
				raise UserWarning('FATAL: Invalid entry on line #%s; invalid Boss address syntax.' % line)
			else:
				continue

		if node.type in SLFTYPE_DOMAIN:
			addr.address['domain'] = bits[1]
			continue


		# it's probably sufficient to have just the keyword and node number, but just to be safe...
		# this should probably be eight (8), but it's not fatal problem if there's no flag field (ie. no comma after the speed)
		if len(bits) < 7:
			raise UserWarning('FATAL: Invalid entry on line #%s; expecting at least seven fields.' % line)


		try:
			node.number = int(bits[1])
		except:
			raise UserWarning('FATAL: Invalid entry on line #%s; node number not numeric' % line)

		if node.type in SLFTYPE_ZONE:
			addr.casZone(node.number)
		elif node.type in SLFTYPE_REGION:
			addr.casRegion(node.number)
		elif node.type in SLFTYPE_NET:
			addr.casNet(node.number)
		elif node.type in SLFTYPE_HUB:
			addr.casHub(node.number)
		elif node.type in SLFTYPE_NODE:
			addr.casNode(node.number)
		elif node.type in SLFTYPE_POINT:
			addr.casPoint(node.number)


		#sys_type, sys_num, sys_name, sys_loc, sys_op, sys_phone, sys_speed[, sys_flags]
		if len(bits) < 8: bits.append(None)  # no flag field will return None
		node.name, node.location, node.sysop, node.phone, node.speed, node.flags = bits[2:8]
		node.bits = bits  # since we still want indexed access...

		yield node



def load_buff(fname, segname):
	# import segment into buffer
	# if the specified file appears to be a ZIP archive, extract most recent segment matching segname (based on file date)
	# returns (buffer, filename, filedate)
	if zipfile.is_zipfile(fname):
		archive = zipfile.ZipFile(fname, 'r')
		#infolist = [ zipinfo for zipinfo in archive.infolist() if (zipinfo.file_size > 0) and (os.path.splitext(os.path.basename(zipinfo.filename))[0].lower() == segname) ]
		infolist = [ zipinfo for zipinfo in archive.infolist() if (zipinfo.file_size > 0) and (fnmatch.fnmatch(os.path.basename(zipinfo.filename).lower(), '%s.[0-9][0-9][0-9]' % segname)) ]
		infolist = sorted(infolist, key=lambda k: k.date_time, reverse=True)

		if infolist:
			if len(infolist) > 1:
				print('WARNING: multiple segments in archive: %s' % ', '.join([zipinfo.filename for zipinfo in infolist]))
			print('Unarchived: %s' % os.path.basename(infolist[0].filename))
			return (archive.read(infolist[0].filename), os.path.basename(infolist[0].filename), time.mktime(list(infolist[0].date_time) + [-1, -1, -1]))
		else:
			raise UserWarning('No "%s" segments in "%s"' % (segname, fname))
	else:
		return (file(fname, 'rb').read(), os.path.basename(fname), os.path.getmtime(fname))


def load_addr(fname):
	# load file containing FTN address masks in RLF, parsed with FTNAddress
	if not os.path.isfile(fname):
		print('File not found: %s' % fname)
		raise SystemExit(1)

	try:
		return [ FTNAddress2.FTNAddress(line.strip()) for line in file(fname, 'rU') ]
	except:
		print('IO error reading file: %s' % fname)
		raise SystemExit(1)


def load_mail(fname):
	# load file containing [success|failure] <from name> <from address> <to name> <to address> <subject> [area] lines
	if not os.path.isfile(fname):
		print('File not found: %s' % fname)
		raise SystemExit(1)

	try:
		import shlex
		recipients = []
		for line in file(fname, 'rU'):
			line = line.strip()
			if (not line) or (line.strip()[:1] in ('#', ';')): continue  # blank or comment
			item = {}
			item['_check_status'], item['orig_name'], item['orig_addr'], item['dest_name'], item['dest_addr'], item['subject'], item['area'] = shlex.split(line.strip())
			recipients.append(item)
	except:
		print('IO error reading file: %s' % fname)
		raise SystemExit(1)

	return recipients


def _send_mail(spool, meta, body):
	if spool is None:
		return

	class ftn(FTNAddress2.FTNAddress):
		def d2(self):
			return '%s/%s' % (self.net, self.node)
		def d3(self):
			return '%s:%s/%s' % (self.zone, self.net, self.node)

	orig_addr = ftn(meta['orig_addr'])
	dest_addr = ftn(meta['dest_addr'])

	pkt = PktWidgets.Type2Plus(DefaultZone=orig_addr.zone)
	pkt.Header['orignet'] = pkt.Header['destnet'] = orig_addr.net
	pkt.Header['orignode'] = pkt.Header['destnode'] = orig_addr.node

	msg = pkt.BlankMessage()
	msg['orignet'], msg['orignode'] = orig_addr.net, orig_addr.node
	msg['destnet'], msg['destnode'] = dest_addr.net, dest_addr.node

	msg['from'] = meta['orig_name']
	msg['to'] = meta['dest_name']
	msg['subject'] = meta['subject']

	msg['body_tail'] = '\r\n\r\n---\r\n'

	if meta['area'] == '':  # netmail
		msg['body_head'] = '\x01INTL %s %s\r\n' % (dest_addr.d3(), orig_addr.d3())

		if (dest_addr.point is not None) and (dest_addr.point > 0):
			msg['body_head'] += '\x01TOPT %s\r\n' % dest_addr.point

		if (orig_addr.point is not None) and (orig_addr.point > 0):
			msg['body_head'] += '\x01FMPT %s\r\n' % orig_addr.point

	else:  # echomail
		msg['body_head'] = 'AREA:%s\r\n' % meta['area']
		msg['body_tail'] += ' * Origin: (%s)\r\n' % orig_addr
		msg['body_tail'] += 'SEEN-BY: %s\r\n' % orig_addr.d2()
		msg['body_tail'] += '\x01PATH: %s\r\n' % orig_addr.d2()

	msg['body_head'] += '\x01MSGID: %s %s\r\n' % (orig_addr, str(uuid.uuid4()).split('-')[0])
	msg['body_head'] += '\x01PID: PktWidgets\r\n'
	#msg['body_head'] += '\x01CHRS: UTF-8 4\r\n'

	msg['body'] = msg['body_head'] + body + msg['body_tail']
	pkt.Messages.append(msg)

	while True:
		fname = os.path.join(spool, str(uuid.uuid4()).split('-')[0] + '.pkt')
		if not os.path.exists(fname):
			try:
				file(fname, 'wb').write(str(pkt))
				break
			except Exception:
				print('Unable to create packet: %s' % fname)
				raise SystemExit(1)


def send_mail(spool, items, fname, status, msg):
	for item in items:
		if status != item['_check_status']:	continue

		body  = 'Hello, %s!\r\n\r\n' % item['dest_name']

		if item['_check_status'] == 'success':
			body += 'This is an automated message to inform you that the segment file "%s" was checked and accepted.\r\n\r\n' % os.path.basename(fname).upper()
			if msg is not None: body += '%s\r\n\r\n' % msg
		elif item['_check_status'] == 'warning':
			body += 'This is an automated message to inform you that the segment file "%s" was checked and accepted, but contained non-critical errors.\r\n\r\n' % os.path.basename(fname).upper()
			if msg is not None: body += 'The error is: %s\r\n\r\n' % msg
		elif item['_check_status'] == 'failure':
			body += 'This is an automated message to inform you that the segment file "%s" was rejected.\r\n\r\n' % os.path.basename(fname).upper()
			if msg is not None: body += 'The error is: %s\r\n\r\n' % msg
			body += 'You can either send a replacement, or do nothing and the previous segment will be used instead.\r\n\r\n'
		else:
			body += 'Something weird happened -- please report this to the operator.\r\n\r\n'

		body += 'Any questions, please contact %s at %s.\r\n' % (item['orig_name'], item['orig_addr'])
		_send_mail(spool, item, body)


class SendAndExit:
	def __init__(self):
		self.exitcode = 1      # exit with this return code
		self.fullname = None   # full path/filename of segment
		self.maillist = None   # from load_mail()
		self.mailpath = None   # path to save PKTs
		self.check_ok = 'failure'  # success, warning, failure

	def __call__(self, msg=None):
		if msg is not None: print(msg)
		if not None in (self.fullname, self.maillist, self.mailpath):
			send_mail(self.mailpath, self.maillist, self.fullname, self.check_ok, msg)
		if self.exitcode is not None:
			raise SystemExit(self.exitcode)


def main():
	print('%(__program__)s v%(__version__)s by %(__author__)s\n' % globals())

	parser = optparse.OptionParser(usage='usage: %prog [options] <segment>')
	parser.add_option('--config', action='store', dest='config', type='str', default=os.path.dirname(sys.argv[0]), metavar='PATH', help='Path to configuration files')
	parser.add_option('--maxage', action='store', dest='maxage', type='int', default=7, metavar='DAYS', help='Maximum segment age in days')
	parser.add_option('--future', action='store', dest='future', type='int', default=7, metavar='DAYS', help='Maximum days segment can be early')
	parser.add_option('--save', action='store', dest='save_path', metavar='PATH', help='Save segment (buffer) to PATH')
	parser.add_option('--spool', action='store', dest='mail_path', metavar='PATH', default=None, help='Save PKT files to PATH')
	opts, args = parser.parse_args()

	if len(args) != 1:
		parser.print_help()
		raise SystemExit(1)

	fullname = os.path.normpath(args[0])
	pathname, filename = os.path.split(fullname)
	segment = os.path.splitext(filename)[0].lower()

	if not os.path.isfile(fullname):
		print('File not found: %s' % fullname)
		raise SystemExit(1)

	print('Processing: %s' % fullname)


	config_files = ('%s.type', '%s.allow', '%s.deny', '%s.req', '%s.warn', '%s.mail')
	config_files = [ os.path.join(opts.config, name % segment) for name in config_files ]

	# config files are given the same base name as the segment name; no config == invalid segment
	for cfg in config_files:
		if not os.path.isfile(cfg):
			print('File not found: %s' % cfg)
			print('Invalid segment name: %s' % segment)
			raise SystemExit(1)
	cfgname_type, cfgname_allow, cfgname_deny, cfgname_req, cfgname_warn, cfgname_mail = config_files

	try:
		cfg_type = file(cfgname_type, 'rU').readline().strip()
	except:
		print('IO error reading file: %s' % cfgname_type)
		raise SystemExit(1)

	if not cfg_type:
		print('Segment type not configured.')
		raise SystemExit(1)

	addr_allow = load_addr(cfgname_allow)
	addr_deny = load_addr(cfgname_deny)
	addr_req = load_addr(cfgname_req)
	addr_warn = load_addr(cfgname_warn)
	mail_list = load_mail(cfgname_mail)

	# preload all the info we'll need to log/netmail success/failure messages
	send_exit = SendAndExit()
	send_exit.fullname = fullname
	send_exit.mailpath = opts.mail_path
	send_exit.maillist = mail_list

	buff, segname, segdate = load_buff(fullname, segment)
	if not buff:
		send_exit('IO error loading segment')

	try:
		srcheader = NodelistWidgets.parse_header(buff)
	except:
		send_exit('Invalid segment header')

	# Type check
	if not srcheader['network'].lower() == cfg_type:
		send_exit('Invalid segment type: %s' % srcheader['network'])

	# CRC check
	srccrc = NodelistWidgets.calc_crc(buff)
	if srccrc != srcheader['crc']:
		send_exit('CRC does not match header.')

	# Age checks
	delta = srcheader['date'].today() - srcheader['date']

	if opts.maxage:
		if delta.days > opts.maxage:
			send_exit('Segment is too old: %s days' % delta.days)

	if opts.future:
		if delta.days < -opts.future:
			send_exit('Segment is too early: %s days' % abs(delta.days))

	# Address check.  Iterate over the nodelist and:
	#	1) check each node is in addr_allow
	#	2) check each node is not in addr_deny
	#	3) remove nodes from addr_req as they are found
	for count, node in enumerate(read_nodelist(buff.split('\r\n'), FTNAddress2.FTNAddress('::::::'))):
		if not node.addr in addr_allow:
			send_exit('Address not allowed: %s' % repr(node.addr))
		if node.addr in addr_deny:
			send_exit('Address prohibited: %s' % repr(node.addr))
		if node.addr in addr_req:
			addr_req.remove(node.addr)
		if node.addr in addr_warn:
			addr_warn.remove(node.addr)

	if addr_req:
		send_exit('Missing nodes (req): %s' % ', '.join([ repr(node) for node in addr_req ]))

	#print('Checked %s nodes.' % count)

	if opts.save_path:
		dname = os.path.normpath(os.path.join(opts.save_path, segname))
		#if not os.path.exists(dname):
		print('Saving to: %s' % dname)
		file(dname, 'wb').write(buff)
		os.utime(dname, (segdate, segdate))
		#else:
		#	print('Output file already exists: %s' % dname)
		#	raise SystemExit(1)

	send_exit.exitcode = 0

	if addr_warn:
		send_exit.check_ok = 'warning'
		send_exit('Missing nodes (warn): %s\nChecked %s nodes.' % (', '.join([ repr(node) for node in addr_warn ]), count))
	else:
		send_exit.check_ok = 'success'
		send_exit('Checked %s nodes.' % count)

if __name__ == '__main__':
	try:
		main()
		#for rec in load_mail('region57.mail'):
		#	send_mail('./', rec, '** TEST **')
	except BaseException as ex:
		if (not hasattr(ex, 'code')) or (ex.code != 0):  # SystemExit(0) == normal exit
			file('fail.sem', 'wb').close()
		raise
